/*
  Diana - A Free Meteorological Visualisation Tool

  Copyright (C) 2017-2018 met.no

  Contact information:
  Norwegian Meteorological Institute
  Box 43 Blindern
  0313 OSLO
  NORWAY
  email: diana@met.no

  This file is part of Diana

  Diana is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Diana is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Diana; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "diObsPlotCluster.h"

#include "diEditManager.h"
#include "diKVListPlotCommand.h"
#include "diObsManager.h"
#include "diObsPlot.h"
#include "diUtilities.h"
#include "util/was_enabled.h"
#include "util/misc_util.h"

#include <puTools/miStringFunctions.h>

#define MILOGGER_CATEGORY "diana.ObsPlotCluster"
#include <miLogger/miLogging.h>

namespace {
const std::string OBS = "OBS";
} // namespace

ObsPlotCluster::ObsPlotCluster(ObsManager* obsm, EditManager* editm)
  : hasDevField_(false)
  , collider_(new ObsPlotCollider)
  , obsm_(obsm)
  , editm_(editm)
{
}

ObsPlotCluster::~ObsPlotCluster()
{
}

const std::string& ObsPlotCluster::plotCommandKey() const
{
  return OBS;
}

void ObsPlotCluster::prepare(const PlotCommand_cpv& inp)
{
  diutil::was_enabled plotenabled;
  for (unsigned int i = 0; i < plots_.size(); i++)
    plotenabled.save(plots_[i]);

  // for now -- erase all obsplots etc..
  //first log stations plotted
  for (size_t i = 0; i < plots_.size(); i++)
    at(i)->logStations();
  cleanup();
  hasDevField_ = false;

  for (PlotCommand_cp pc : inp) {
    if (KVListPlotCommand_cp cmd = std::dynamic_pointer_cast<const KVListPlotCommand>(pc)) {
      const std::string dialogname = ObsPlot::extractDialogname(cmd->all());
      if (dialogname.empty()) {
        METLIBS_LOG_WARN("probably malformed observation plot specification");
        continue;
      }
      const ObsPlotType plottype = ObsPlot::extractPlottype(dialogname);
      std::unique_ptr<ObsPlot> op(new ObsPlot(dialogname, plottype));
      obsm_->setPlotDefaults(op.get());
      op->setPlotInfo(cmd->all());
      op->setCanvas(canvas_);
      op->setCollider(collider_.get());
      hasDevField_ |= op->mslp();

      plotenabled.restore(op.get());
      plots_.push_back(op.release());
    }
  }
  collider_->clear();
}

bool ObsPlotCluster::update(bool ifNeeded, const miutil::miTime& t)
{
  bool havedata = false;

  if (!ifNeeded) {
    for (size_t i = 0; i < plots_.size(); i++)
      at(i)->logStations();
  }
  for (size_t i = 0; i < plots_.size(); i++) {
    ObsPlot* op = at(i);
    if (!ifNeeded || obsm_->updateTimes(op)) {
      if (obsm_->prepare(op, t)) {
        havedata = true;
      }
    }
    //update list of positions ( used in "PPPP-mslp")
    // TODO this is kind of prepares changeProjection of all ObsPlot's with mslp() == true, to be used in EditManager::interpolateEditFields
    op->updateObsPositions();
  }
  return havedata;
}

void ObsPlotCluster::plot(DiGLPainter* gl, PlotOrder zorder)
{
  if (zorder != PO_LINES && zorder != PO_OVERLAY)
    return;

  collider_->clear();

  // plot observations (if in fieldEditMode  and the option obs_mslp is true, plot observations in overlay)

  const bool obsedit = (hasDevField_ && editm_->isObsEdit());
  const bool plotoverlay = (zorder == PO_OVERLAY && obsedit);
  const bool plotunderlay = (zorder == PO_LINES && !obsedit);

  if (plotoverlay) {
    for (size_t i = 0; i < plots_.size(); i++) {
      ObsPlot* op = at(i);
      if (editm_->interpolateEditField(op->getObsPositions()))
        op->updateFromEditField();
    }
  }
  if (plotunderlay || plotoverlay)
    PlotCluster::plot(gl, zorder);
}

void ObsPlotCluster::getDataAnnotations(std::vector<std::string>& anno) const
{
  for (Plot* p : plots_)
    static_cast<ObsPlot*>(p)->getDataAnnotations(anno);
}

void ObsPlotCluster::getExtraAnnotations(std::vector<AnnotationPlot*>& vap)
{
  //get obs annotations
  for (Plot* p : plots_) {
    ObsPlot* op = static_cast<ObsPlot*>(p);
    if (!op->isEnabled())
      continue;
    PlotCommand_cpv obsinfo = op->getObsExtraAnnotations();
    for (PlotCommand_cp pc : obsinfo) {
      AnnotationPlot* ap = new AnnotationPlot(pc);
      vap.push_back(ap);
    }
  }
}

plottimes_t ObsPlotCluster::getTimes()
{
  std::set<std::string> readernames;
  for (Plot* p : plots_)
    diutil::insert_all(readernames, static_cast<const ObsPlot*>(p)->readerNames());
  if (!readernames.empty()) {
    return obsm_->getObsTimes(readernames);
  } else {
    return plottimes_t();
  }
}

const std::string& ObsPlotCluster::keyPlotElement() const
{
  return OBS;
}

bool ObsPlotCluster::findObs(int x, int y)
{
  bool found = false;

  for (size_t i = 0; i < plots_.size(); i++)
    if (at(i)->showpos_findObs(x, y))
      found = true;

  return found;
}

std::string ObsPlotCluster::getObsPopupText(int x, int y)
{
  size_t n = plots_.size();
  std::string obsText = "";

  for (size_t i = 0; i < n; i++)
    if (at(i)->getObsPopupText(x, y, obsText))
      return obsText;

  return obsText;
}

void ObsPlotCluster::nextObs(bool next)
{
  for (size_t i = 0; i < plots_.size(); i++)
    at(i)->nextObs(next);
}

std::vector<ObsPlot*> ObsPlotCluster::getObsPlots() const
{
  std::vector<ObsPlot*> obsplots;
  obsplots.reserve(plots_.size());
  for (std::vector<Plot*>::const_iterator it = plots_.begin(); it != plots_.end(); ++it)
    obsplots.push_back(static_cast<ObsPlot*>(*it));
  return obsplots;
}

ObsPlot* ObsPlotCluster::at(size_t i) const
{
  return static_cast<ObsPlot*>(plots_[i]);
}
